import portion as P

from .configuration import Variable
from .util import *


class LiftedSet(Variable):
    """
    Represents a lifted subset of the universe/a set

    Attributes
    ----------
    name : str
        not important at the moment
    size : SizeFormula
        describes the set of cardinality values that are valid for this set
    ccs : [CCountings]
        describes properties of the set in terms of counting formulas
    relevant : [CCounting]
        defines which groups of indistinguishable objects need to be accounted for
    source : SetFormula
        the set of which the LiftedSet is subset
    """

    def __init__(self, universe, size, ccs=[]):
        """
        A LiftedSet is a set of constraints restricting the powerset of the universe.

        Args:
            universe (Multiset): universe of the problem
            size (CSize): constrains the size of the set (at least non empty)
            ccs ([CConstraints], optional): further counting constraints. Defaults to [].
        """
        super().__init__(universe, f"part. of {universe}")
        self.universe = universe
        self.size = size
        self.ccs = self.compact_ccs(ccs)
        self.histogram = {}
        self.check_bound()

    def __and__(self, rhs):
        """
        Conjoin with another set of constraints

        Args:
            rhs (LiftedSet): other

        Returns:
            LiftedSet: conjunction of constraints
        """
        # name = f"{self.name} /\ {rhs.name}"
        size = self.size & rhs.size
        ccs = self.ccs + rhs.ccs
        ccs = self.compact_ccs(ccs)
        return LiftedSet(self.universe, size, ccs)

    def __contains__(self, constr):
        """
        Check if a constraint is present in the lifted set

        Args:
            constr (Constraint): Counting or CSize

        Returns:
            bool: true if there is a constraint implying constr
        """
        if hasattr(constr, "formula"):
            for cc in self.ccs:
                if cc.formula == constr.formula:
                    return cc.values in constr.values
            return False
        else:
            return self.size.values in constr.values

    def __eq__(self, rhs):
        """
        Check if two lifted sets have the same constraints

        Args:
            rhs (LiftedSet): other

        Returns:
            bool:
        """
        if self.size != rhs.size:
            return False
        elif self.ccs != rhs.ccs:
            return False
        elif self.histogram != rhs.histogram:
            return False
        else:
            return True

    __hash__ = Variable.__hash__

    def __str__(self):
        s = f"{self.name}: {self.size}"
        if len(self.ccs) > 0:
            s += "\n"
            s += "\t counting: "
            s += ",".join([str(c) for c in self.ccs])
        if len(self.histogram) > 0:
            s += "\n\t histogram:" + str(self.histogram)
        return s

    def add_cc(self, cc):
        """
        Add a new counting constraint replacing infs with int numbers

        Args:
            cc (CConstraint): new counting constraint

        Returns:
            LiftedSet: return a new lifted set wit the additional constraint
        """
        s = self.size.values.upper
        s = s - 1 if self.size.values.right == P.OPEN else s
        if cc.values.upper == P.inf:
            safe_vals = cc.values.replace(upper=s + 1)
        else:
            safe_vals = cc.values
        if cc.values.lower == P.inf:
            safe_vals = safe_vals.replace(lower=s)
        cc.values = safe_vals
        ccs = self.ccs + [cc]
        return LiftedSet(self.universe, self.size, ccs)

    def bound(self, rv_set):
        """
        Check if the number of entities from a relevant set is determined by constraints

        Args:
            rv_set (Multiset): relevant set

        Returns:
            bool: wether the number of entities from rv_set can be inferred from the constraints
        """
        bound = False
        if rv_set == self.universe:
            bound = is_singleton(self.size)
        else:
            for cc in self.ccs:
                if cc.formula == rv_set and is_singleton(cc.values):
                    self.histogram[rv_set] = cc.values.lower
                else:
                    bound = bound or is_singleton(cc.values)
        return bound

    def compact_ccs(self, counts):
        """
        Clean the list of counting constraints removing trivial/duplicates

        Args:
            counts ([CConstraints]): list of counting constraints

        Returns:
            [CConstraints]: compacted list
        """
        compact = []
        remove = []
        indexes = range(0, len(counts))
        for i in indexes:
            cc1 = counts[i]
            for j in [j for j in indexes if j > i]:
                cc2 = counts[j]
                if cc1 != cc2:
                    if cc1.formula == cc2.formula:
                        new_interval = cc1.values & cc2.values
                        merged_cc = cc1.copy()
                        merged_cc.values = new_interval
                        compact.append(merged_cc)
                        remove += [i, j]
                else:  # cc1 == cc2:
                    remove += [i]
        keep = [i for i in indexes if not i in remove]
        compact += [counts[i] for i in keep]
        final = len(remove) == 0
        result = compact if final else self.compact_ccs(compact)
        return result

    def check_bound(self):
        """
        Removes inf upper bound from counting constraints
        """
        for cc in self.ccs:
            if cc.values.upper == P.inf:
                ub = self.size.values.upper + 1
                max_int = P.closed(0, ub)
                cc.values = cc.values & max_int

    def copy(self):
        size = self.size.copy()
        ccs = self.ccs.copy()
        hist = self.histogram.copy()
        ls = LiftedSet(self.universe, size, ccs)
        ls.histogram = hist
        return ls

    def disjoint(self, constr):
        """
        Check if a constraint is incompatible with current constraints

        Args:
            constr (Constraint): CConstraint or CSize

        Returns:
            bool: test result
        """
        if hasattr(constr, "formula"):  # counting formula
            dis_ccs = False
            for cc in self.ccs:
                if cc.formula == constr.formula:
                    if cc.values & constr.values == P.empty():
                        dis_ccs = True
            return dis_ccs
        else:  # size formula
            disj = (self.size.values & constr.values) == P.empty()
            return disj

    def feasible(self, rv_set, n, n_class=1):
        """Check if there is some incompatible constraint with the distribution of n entities of
        rv_set over n_class exchangeable variables

        Args:
            rv_set (SetFormula): the domain formula being distributed
            n (int): number of entities
            n_class (int, optional): number of exchangeable variables together with self. Defaults to 1.

        Returns:
            Boolean: True if no constraint is violated by the distribution
        """
        # print("-----")
        # print(self, rv_set, n, n_class)
        if rv_set == self.universe:
            # we are distributing the universe: a size constraint
            size_class = self.size.values.replace(
                upper=lambda v: n_class * v, lower=lambda v: n_class * v
            )
            feasible = n in size_class
            # if not feasible:
            # print(f"Unfeasible size: {n} not in {size_class}")
            return feasible
        else:
            for cc in self.ccs:
                if rv_set in cc.formula:
                    cc_class = cc.values.replace(
                        upper=lambda v: n_class * v, lower=lambda v: n_class * v
                    )
                    # we would have too many elements for a counting constraint
                    if cc_class.empty:
                        return False
                    lb = (
                        cc_class.lower
                        if cc_class.left == P.CLOSED
                        else cc_class.lower + 1
                    )
                    ub = (
                        cc_class.upper
                        if cc_class.right == P.CLOSED
                        else cc_class.upper - 1
                    )
                    feasible = n <= ub
                    if not feasible:
                        # print(f"Unfeasible because {n} {rv_set} > {ub} {cc.formula}")
                        return feasible
                    # now check histograms: because rv_set in cc.formula we start with
                    # n entities that satisfy cc.formula
                    n_cc = 0
                    unfixed = []
                    # then we count which entities are fixed that satisfy cc.formula
                    for rvs in self.histogram:
                        if rvs in cc.formula and rvs != rv_set:
                            if self.histogram[rvs] == -1:
                                unfixed.append(rvs)
                            else:
                                n_cc += self.histogram[rvs]
                    if len(unfixed) == 0:
                        # if everything is fixed then the constraint should be satisfied
                        if n_cc + n not in cc_class:
                            # print(f"Everything related to {cc} is fixed but still unsat")
                            return False
                    else:
                        # if the sizes of the unfixed (in cc.formula) are too small to satisfy
                        # a lower bound of the constraint then unsat
                        sizes = sum([rvs.size() for rvs in unfixed])
                        if (sizes + n_cc + n) * n_class < lb:
                            # print(f"{sizes} entities from other relevant sets and {n_class} variables are not enough for {cc}")
                            return False
            if len(self.histogram) > 1:
                # check lasts unfixed value for histogram: everything not yet sat should become sat
                vals = [n for n in self.histogram.values() if n > -1]
                if (
                    len(vals) == len(self.histogram) - 1
                    and self.histogram[rv_set] == -1
                ):  # last relevant set: check size
                    fixed = sum(vals)  # respect overall size
                    if fixed + (n / n_class) not in self.size.values:
                        # print(f"{rv_set} is the last to be fixed but {fixed+(n/n_class)} is not a valid size")
                        return False
            return True

    def set_labels(self, labels):
        super().set_labels(labels)
        for cc in self.ccs:
            cc.set_labels(labels)

    def rv_size(self, relevant):
        """
        Return the (possible) number of entities from a specific relevant subset

        Args:
            relevant (SetFormula): a relevant set

        Returns:
            portion: interval of allowed size of the subset of entities in relevant
        """
        for cc in self.ccs:
            if cc.formula == relevant:
                return cc.values

    def size_is_defined(self):
        """
        Check if the possible sizes are a unique value

        Returns:
            bool: result
        """
        s = 0
        for e in self.size.values:
            if not e.empty:
                if e.left == P.CLOSED and e.right == P.CLOSED:
                    s += e.upper - e.lower + 1
                elif e.left == P.OPEN and e.right == P.OPEN:
                    s += e.upper - e.lower - 1
                else:
                    s += e.upper - e.lower
        return s == 1

    def relevant(self):
        """
        Returns the sets that are relevant for this LiftedSet

        Returns:
            {Multiset}: the relevant sets
        """
        relevant = set([cc.formula for cc in self.ccs])
        if len(relevant) == 0:
            relevant = {self.universe}
        return relevant

    def satisfies(self, constraint, lb, ub):
        """
        Check if a constraint is already implied by others

        Args:
            constraint (CConstraint): constraint to be checked
            lb (int): lower bound on how many sets have to satisfy the level 1 property
            ub (int): upper bound on how many sets have to satisfy the level 1 property

        Returns:
            bool: result
        """
        sat = None
        for cc in self.ccs:
            if cc.formula in constraint.formula:
                int = constraint.values & cc.values
                if cc.values in constraint.values:
                    sat = True
                elif int.empty:
                    sat = False
        s = constraint.formula.size()
        cc_lb, cc_ub = interval_closed(
            constraint.values, ub_default=self.size.values.upper
        )
        if cc_lb > cc_ub:
            # asking to have more elements than the size of the part
            sat = False
        if cc_lb * lb > s:
            # asking to distribute more elements than available
            sat = False
        return sat



from portion.dict import IntervalDict

from .level_1 import SetFormula
from .util import *

class Union(object):
    def __init__(self, formula, size=None, subsets=[]):
        self.formula = formula
        self.size = size
        self.subsets = subsets

    def __repr__(self):
        descr = f"{self.formula}: {self.size}\n"
        for s in self.subsets:
            descr += f"\t{s}\n"
        return descr


class Venn(object):
    """
    A class to represent a Venn diagram: the universe is divided into subsets
    associated to the number of objects belonging to the subsets.

    """

    def __init__(self):
        self.base_sets = []
        self.base_indist = {}
        self.indist = {}
        self.sets = {}
        self.parts = {}

    def __repr__(self):
        descr = ""
        for set in self.sets:
            descr += f"#({self.area2formula(set)}) = {self.sets[set]}\n"
        return descr

    def add_base_set(self, set, dist):
        """
        Add a set which is a label and not a formula

        Args:
            set (str): label
            dist (bool): the elements are labelled
        """
        self.base_sets.append(set)
        self.base_indist[set] = not dist

    def add_set(self, set, size):
        dnfied = dnfy(set)
        if isinstance(dnfied, Or):
            subsets = [s for s in flatten(Or, dnfied)]
        else:
            subsets = []
        u = Union(dnfied, size, subsets)
        self.sets[set] = u

    def subsets(self):
        """
        For all base sets collect subsets from declared intersections
        """
        sets = [self.formula2area(b) for b in self.base_sets]
        for set in sets:
            subs = self.get_subsets(set)
            f = self.sets.get(self.area2formula(set))
            if f is not None:
                s = f.size
            else:
                s = None
            name = self.area2formula(set)
            if len(subs) > 0:
                if name not in self.sets:
                    self.sets[name] = Union(name, s, subs)
                else:
                    self.sets[name].subsets += subs
            else:
                self.sets[name] = Union(name, s, [])
                # self.parts.append(set)

    def get_subsets(self, set1):
        """
        Check which of the sets with declared size are subsets of set 1
        """
        declared = [self.formula2area(k) for k in self.sets.keys()]
        subsets = set()
        for set2 in declared:
            if set1 != set2:
                if self.is_subset(set2, set1):
                    subsets2 = self.get_subsets(set2)
                    if len(subsets2) > 0:
                        subsets.union(subsets2)
                    else:
                        subsets.add(set2)
        return subsets

    def build_parts(self):
        univ_area = (0,) * len(self.base_sets)
        parts = self.get_parts(univ_area)
        for p in parts:
            f = self.area2formula(p)
            if f in self.sets:
                self.parts[p] = self.sets[f].size
        return parts

    def is_subset(self, set1, set2):
        """
        Checks if set1 is contained in set2 by comparing codes

        Args:
            set1 (tuple(int)): code of set 1
            set2 (tuple(int)): code of set 2

        Returns:
            bool: set1 in set2
        """
        subset = True
        s1 = self.pad(set1)
        s2 = self.pad(set2)
        for i in range(0, len(self.base_sets)):
            if s1[i] != 0 and s1[i] == -s2[i]:
                subset = False
            if s1[i] == 0 and s2[i] == 1:
                subset = False
        return subset

    def infer_union_size(self, u):
        """
        Inference on the sizes of the sets

        Args:
            u (Union): a set

        Raises:
            Exception: we might find inconsistent sizes
            Exception: we might miss some sizes
        """
        f = self.formula2area(u.formula)
        parts = [tuple(p) for p in self.get_parts(f)]
        known = [p for p in parts if p in self.parts]
        ksize = sum([self.parts[part] for part in known])
        unknown = [p for p in parts if p not in self.parts]
        if len(unknown) == 0:
            if u.size is None:
                u.size = ksize
            elif u.size != ksize:
                raise Exception(
                    f"Inconsistent size of {u.formula} (expected {u.size}, got {ksize})"
                )
        elif len(unknown) == 1:
            if u.size is None:
                u.size = ksize
                self.parts[unknown[0]] = 0
            else:
                self.parts[unknown[0]] = u.size - ksize
        else:
            if u.size is None:
                if len(known) == 0:
                    raise Exception(
                        f"I don't know neither the size of {u.formula} nor any of its subsets"
                    )
                else:
                    u.size = ksize
            else:
                if len(known) == 0:
                    self.parts[f] = u.size
                # missing the case where both #A and #B are unknown but #A u B is known

    def infer_indistinguishability(self):
        for p in self.parts:
            indist = self.parts[p] > 1
            sets = [self.base_sets[i] for i, s in enumerate(p) if s == 1]
            for set in sets:
                indist = indist and self.base_indist[set]
            self.indist[p] = indist

    def infer(self):
        self.subsets()
        self.build_parts()
        for set in self.sets:
            self.infer_union_size(self.sets[set])
        self.infer_indistinguishability()

    def area2formula(self, area):
        """
        Convert an area code int

        Args:
            area (tuple(-1/1/0)): whether a base set is excluded/included/neither

        Returns:
            And: intersection of base sets or their negation according to the code
        """
        bsets = []
        for i, n in enumerate(area):
            if n == 1:
                bsets.append(self.base_sets[i])
            elif n == -1:
                bsets.append(Not(self.base_sets[i]))
            else:
                pass
        return nest(And, bsets)

    def formula2area(self, formula):
        part = [0] * len(self.base_sets)
        sets = flatten(And, formula)
        for s in sets:
            negated = isinstance(s, Not)
            if negated:
                i = self.base_sets.index(s.child)
                part[i] = -1
            else:
                i = self.base_sets.index(s)
                part[i] = 1
        return tuple(part)

    def pad(self, area):
        """
        Add zeros to a set description created before adding all base sets

        Args:
            area (tuple(int)): the partial encoding
        Returns:
            tuple(int): same encoding with additional zeros
        """
        padded = []
        for id in area:
            padded.append(id)
        pad = len(self.base_sets) - len(area)
        padded += [0] * pad
        return tuple(padded)

    def get_parts(self, area):
        """
        Given a region code of the Venn diagram returns all the partition
        codes that the area covers

        Args:
            area (tuple(int)): code of the set

        Returns:
            [tuple(int)]: list of parts in the set
        """
        if len(area) == 0:
            return [()]
        else:
            try:
                z = area.index(0)
                p1 = [area[:z] + (1,) + r for r in self.get_parts(area[z + 1 :])]
                p2 = [area[:z] + (-1,) + r for r in self.get_parts(area[z + 1 :])]
                return p1 + p2
            except ValueError:
                return [area]

    def update_domains(self, problem):
        """
        Convert a Venn diagram into domains with the correct intersections

        Args:
            problem (Problem): problem corresponding to the Venn diagram

        Returns:
            Problem: a problem with the entities represented explicitly
        """
        self.infer()
        intervals = {}
        for part in self.parts:
            lab = str(self.area2formula(part))
            lab = lab.replace(" ", "").replace("(", "").replace(")", "")
            n = self.parts[part]
            if self.indist[part]:
                e_list = [str(lab)] * n
            else:
                e_list = []
                for j in range(0, n):
                    e = f"{lab}_{j}"
                    e_list.append(e)
            p_int = list2interval(problem, e_list, True)
            intervals[part] = p_int
        for set in self.base_sets:
            area = self.formula2area(set)
            if area in self.parts:
                subsets = [intervals[area]]
            else:
                subsets = [intervals[p] for p in self.get_parts(area)]
            entities = IntervalDict()
            for sub in subsets:
                entities = entities.combine(sub, how=is_distinguishable)
            d = SetFormula(set, entities, problem.universe)
            problem.add_domain(d)
        problem.compute_universe()
        return problem
